@@ -1,9 +1,9 @@
+use anyhow::{ensure, Result};
 use crc32c::*;
 use fs2::FileExt;
 use std::cmp::Ordering;
 use std::convert::TryInto;
 use std::fs::{File, OpenOptions};
-use std::io::Result;
 use std::iter;
 use std::ops::Bound::*;
 use std::ops::{Bound, RangeBounds};
@@ -31,7 +31,7 @@ pub type Value = Vec<u8>;
 const PAGE_SIZE: usize = 8192;
 const MAGIC: u32 = 0xBACE2021;
 const VERSION: u32 = 1;
-const METADATA_SIZE: usize = 6 * 4;
+const METADATA_SIZE: usize = 7 * 4;
 const PAGE_HEADER_SIZE: usize = 2; // now page header contains just number of items in the page
 const MAX_VALUE_LEN: usize = PAGE_SIZE / 4; // assume that pages may fit at least 3 items
 const MAX_KEY_LEN: usize = u8::MAX as usize; // should fit in one byte
@@ -51,7 +51,7 @@ enum LookupOp<'a> {
     GreaterOrEqual(&'a Key),
 }
 
-#[derive(PartialEq)]
+#[derive(PartialEq, Copy, Clone, Debug)]
 enum DatabaseState {
     InRecovery,
     Opened,
@@ -59,6 +59,12 @@ enum DatabaseState {
     Corrupted,
 }
 
+impl Default for DatabaseState {
+    fn default() -> Self {
+        Self::Closed
+    }
+}
+
 #[derive(PartialEq)]
 enum AccessMode {
     ReadOnly,
@@ -99,6 +105,31 @@ pub struct RecoveryStatus {
     pub recovery_end: u64,
 }
 
+///
+/// Database info
+///
+#[derive(Clone, Copy, Debug, Default)]
+pub struct DatabaseInfo {
+    /// Number of pinned pages in buffer cache
+    n_pinned_pages: usize,
+    /// Total number of pages in buffer cache
+    n_cached_pages: usize,
+    /// Heaight of B-Tree
+    tree_height: usize,
+    /// Size of database file
+    db_size: u64,
+    /// Total size of used pages
+    db_used: u64,
+    /// Current WAL size
+    log_size: u64,
+    /// number of committed transactions in this session
+    n_committed_transactions: u64,
+    /// number of aborted transactions in this session
+    n_aborted_transactions: u64,
+    /// State of the database
+    state: DatabaseState,
+}
+
 ///
 /// Abstract storage bidirectional iterator
 ///
@@ -301,6 +332,7 @@ struct Metadata {
     version: u32, // storage format version
     free: PageId, // L1 list of free pages
     size: PageId, // size of database (pages)
+    used: PageId, // numer of used database pages
     root: PageId, // B-Tree root page
     height: u32,  // height of B-Tree
 }
@@ -324,6 +356,7 @@ impl Metadata {
 struct Database {
     meta: Metadata,           // cached metadata (stored in root page)
     lsn: LSN,                 // database modification counter
+    n_aborted_txns: LSN,      // number of aborted transactions
     state: DatabaseState,     // database state
     wal_pos: u64,             // current opsition in log file
     recovery: RecoveryStatus, // status of recovery
@@ -343,10 +376,12 @@ struct BufferManager {
 
     free_pages: BufferId,  // L1-list of free pages
     dirty_pages: BufferId, // L1-list of dirty pages
-    used: BufferId,        // amount of used pages in the cache
+    used: BufferId,        // used part of page pool
+    pinned: BufferId,      // amount of pinned pages
+    cached: BufferId,      // amount of cached pages
 
     hash_table: Vec<BufferId>, // array containing indexes of collision chains
-    pages: Vec<PageHeader>,    //page data
+    pages: Vec<PageHeader>,    // page data
 }
 
 //
@@ -564,6 +599,7 @@ impl BufferManager {
         self.pages[id as usize].access_count = 0;
         self.pages[id as usize].next = self.head;
         self.pages[id as usize].prev = 0;
+        self.pinned -= 1;
         if self.head != 0 {
             self.pages[self.head as usize].prev = id;
         } else {
@@ -589,6 +625,7 @@ impl BufferManager {
         } else {
             self.pages[next].prev = self.pages[id as usize].prev;
         }
+        self.pinned += 1;
     }
 
     //
@@ -623,6 +660,7 @@ impl BufferManager {
         self.remove(id);
         self.pages[id as usize].next = self.free_pages;
         self.free_pages = id;
+        self.cached -= 1;
     }
 
     //
@@ -654,7 +692,7 @@ impl BufferManager {
     //
     // Find buffer with specified page or allocate new buffer
     //
-    fn get_buffer(&mut self, pid: PageId) -> BufferId {
+    fn get_buffer(&mut self, pid: PageId) -> Result<BufferId> {
         let hash = pid as usize % self.hash_table.len();
         let mut h = self.hash_table[hash];
         while h != 0 {
@@ -665,7 +703,7 @@ impl BufferManager {
                     self.pin(h);
                 }
                 self.pages[h as usize].access_count = access_count + 1;
-                return h;
+                return Ok(h);
             }
             h = self.pages[h as usize].collision;
         }
@@ -674,14 +712,18 @@ impl BufferManager {
         if h != 0 {
             // has some free pages
             self.free_pages = self.pages[h as usize].next;
+            self.cached += 1;
+            self.pinned += 1;
         } else {
             h = self.used;
             if (h as usize) < self.hash_table.len() {
                 self.used += 1;
+                self.cached += 1;
+                self.pinned += 1;
             } else {
                 // Replace least recently used page
                 let victim = self.tail;
-                assert!(victim != 0);
+                ensure!(victim != 0);
                 debug_assert!(self.pages[victim as usize].access_count == 0);
                 debug_assert!((self.pages[victim as usize].state & PAGE_DIRTY) == 0);
                 self.pin(victim);
@@ -693,7 +735,7 @@ impl BufferManager {
         self.pages[h as usize].id = pid;
         self.pages[h as usize].state = PAGE_RAW;
         self.insert(h);
-        h
+        Ok(h)
     }
 }
 
@@ -724,25 +766,26 @@ impl Storage {
     //
     // Allocate new page in storage and get buffer for it
     //
-    fn new_page(&self, db: &mut RwLockWriteGuard<Database>) -> PageGuard<'_> {
+    fn new_page(&self, db: &mut RwLockWriteGuard<Database>) -> Result<PageGuard<'_>> {
         let mut bm = self.buf_mgr.lock().unwrap();
         let free = db.meta.free;
         let buf;
         if free != 0 {
-            buf = bm.get_buffer(free);
+            buf = bm.get_buffer(free)?;
             let page = self.pool[buf as usize].read().unwrap();
             db.meta.free = page.get_u32(0);
         } else {
             // extend storage
-            buf = bm.get_buffer(db.meta.size);
+            buf = bm.get_buffer(db.meta.size)?;
             db.meta.size += 1;
         }
+        db.meta.used += 1;
         bm.modify_buffer(buf);
-        PageGuard {
+        Ok(PageGuard {
             buf,
             pid: bm.pages[buf as usize].id,
             storage: &self,
-        }
+        })
     }
 
     //
@@ -751,7 +794,7 @@ impl Storage {
     //
     fn get_page(&self, pid: PageId, mode: AccessMode) -> Result<PageGuard<'_>> {
         let mut bm = self.buf_mgr.lock().unwrap();
-        let buf = bm.get_buffer(pid);
+        let buf = bm.get_buffer(pid)?;
         if (bm.pages[buf as usize].state & PAGE_BUSY) != 0 {
             // Some other thread is loading buffer: just wait until it done
             bm.pages[buf as usize].state |= PAGE_WAIT;
@@ -912,6 +955,7 @@ impl Storage {
             self.file.read_exact_at(&mut page.data, 0)?;
             db.meta = Metadata::unpack(&page.data);
         }
+        db.n_aborted_txns += 1;
         Ok(())
     }
 
@@ -934,7 +978,7 @@ impl Storage {
             file.try_lock_exclusive()?;
             file.read_exact_at(&mut buf, 0)?;
             let meta = Metadata::unpack(&buf);
-            assert!(meta.magic == MAGIC && meta.version == VERSION && meta.size >= 1);
+            ensure!(meta.magic == MAGIC && meta.version == VERSION && meta.size >= 1);
             (file, meta)
         } else {
             let file = OpenOptions::new()
@@ -949,6 +993,7 @@ impl Storage {
                 version: VERSION,
                 free: 0,
                 size: 1,
+                used: 1,
                 root: 0,
                 height: 0,
             };
@@ -976,6 +1021,8 @@ impl Storage {
                 free_pages: 0,
                 dirty_pages: 0,
                 used: 1, // pinned root page
+                cached: 1,
+                pinned: 1,
                 hash_table: vec![0; cache_size],
                 pages: vec![PageHeader::new(); cache_size],
             }),
@@ -987,6 +1034,7 @@ impl Storage {
             checkpoint_interval,
             db: RwLock::new(Database {
                 lsn: 0,
+                n_aborted_txns: 0,
                 meta,
                 recovery: RecoveryStatus {
                     ..Default::default()
@@ -1098,12 +1146,12 @@ impl Storage {
         db: &mut RwLockWriteGuard<Database>,
         key: &Key,
         value: &Value,
-    ) -> PageId {
-        let pin = self.new_page(db);
+    ) -> Result<PageId> {
+        let pin = self.new_page(db)?;
         let mut page = self.pool[pin.buf as usize].write().unwrap();
         page.set_n_items(0);
         page.insert_item(0, key, value);
-        pin.pid
+        Ok(pin.pid)
     }
 
     //
@@ -1115,13 +1163,15 @@ impl Storage {
         key: &Key,
         left_child: PageId,
         right_child: PageId,
-    ) -> PageId {
-        let pin = self.new_page(db);
+    ) -> Result<PageId> {
+        let pin = self.new_page(db)?;
         let mut page = self.pool[pin.buf as usize].write().unwrap();
         page.set_n_items(0);
+        debug_assert!(left_child != 0);
+        debug_assert!(right_child != 0);
         page.insert_item(0, key, &left_child.to_be_bytes().to_vec());
         page.insert_item(1, &vec![], &right_child.to_be_bytes().to_vec());
-        pin.pid
+        Ok(pin.pid)
     }
 
     //
@@ -1136,10 +1186,10 @@ impl Storage {
         ip: ItemPointer,
         key: &Key,
         value: &Value,
-    ) -> Option<(Key, PageId)> {
+    ) -> Result<Option<(Key, PageId)>> {
         if !page.insert_item(ip, key, value) {
             // page is full then divide page
-            let pin = self.new_page(db);
+            let pin = self.new_page(db)?;
             let mut new_page = self.pool[pin.buf as usize].write().unwrap();
             let split = page.split(&mut new_page);
             let ok = if ip > split {
@@ -1147,10 +1197,10 @@ impl Storage {
             } else {
                 new_page.insert_item(ip, key, value)
             };
-            assert!(ok);
-            Some((new_page.get_last_key(), pin.pid))
+            ensure!(ok);
+            Ok(Some((new_page.get_last_key(), pin.pid)))
         } else {
-            None
+            Ok(None)
         }
     }
 
@@ -1199,6 +1249,7 @@ impl Storage {
             // free page
             page.set_u32(0, db.meta.free);
             db.meta.free = pid;
+            db.meta.used -= 1;
             Ok(true)
         } else {
             Ok(false)
@@ -1237,7 +1288,7 @@ impl Storage {
                 // replace old value with new one: just remove old one and reinsert new key-value pair
                 page.remove_key(r);
             }
-            Ok(self.btree_insert_in_page(db, &mut page, r, key, value))
+            self.btree_insert_in_page(db, &mut page, r, key, value)
         } else {
             // recurse to next level
             debug_assert!(r < n);
@@ -1245,15 +1296,8 @@ impl Storage {
             if let Some((key, child)) = overflow {
                 // insert new page before original
                 self.modify_page(pin.buf);
-                Ok(
-                    self.btree_insert_in_page(
-                        db,
-                        &mut page,
-                        r,
-                        &key,
-                        &child.to_be_bytes().to_vec(),
-                    ),
-                )
+                debug_assert!(child != 0);
+                self.btree_insert_in_page(db, &mut page, r, &key, &child.to_be_bytes().to_vec())
             } else {
                 Ok(None)
             }
@@ -1269,15 +1313,15 @@ impl Storage {
         key: &Key,
         value: &Value,
     ) -> Result<()> {
-        assert!(key.len() != 0 && key.len() <= MAX_KEY_LEN && value.len() <= MAX_VALUE_LEN);
+        ensure!(key.len() != 0 && key.len() <= MAX_KEY_LEN && value.len() <= MAX_VALUE_LEN);
         if db.meta.root == 0 {
-            db.meta.root = self.btree_allocate_leaf_page(db, key, value);
+            db.meta.root = self.btree_allocate_leaf_page(db, key, value)?;
             db.meta.height = 1;
         } else if let Some((key, page)) =
             self.btree_insert(db, db.meta.root, key, value, db.meta.height)?
         {
             // overflow
-            db.meta.root = self.btree_allocate_internal_page(db, &key, page, db.meta.root);
+            db.meta.root = self.btree_allocate_internal_page(db, &key, page, db.meta.root)?;
             db.meta.height += 1;
         }
         Ok(())
@@ -1307,6 +1351,7 @@ impl Storage {
         op: LookupOp,
         path: &mut TreePath,
     ) -> Result<Option<(Key, Value)>> {
+        ensure!(db.state == DatabaseState::Opened);
         match op {
             LookupOp::First => {
                 // Locate left-most element in the tree
@@ -1372,7 +1417,6 @@ impl Storage {
     // Perform lookup in the database. Initialize path in the tree current element or reset path if no element is found or end of set is reached.
     //
     fn lookup(&self, db: &Database, op: LookupOp, path: &mut TreePath) {
-        assert!(db.state == DatabaseState::Opened);
         let result = self.do_lookup(db, op, path);
         if result.is_err() {
             path.curr = None;
@@ -1413,6 +1457,8 @@ impl Storage {
                 Ok(false)
             }
         } else {
+            debug_assert!(r < n);
+            debug_assert!(page.get_child(r) != 0);
             self.find(page.get_child(r), path, key, height - 1)
         }
     }
@@ -1422,6 +1468,7 @@ impl Storage {
     // Returns true is such element is found and path is successfully reconstructed, false otherwise.
     //
     fn reconstruct_path(&self, path: &mut TreePath, db: &Database) -> Result<bool> {
+        path.stack.clear();
         if let Some((key, _value)) = &path.curr.clone() {
             if self.find(db.meta.root, path, &key, db.meta.height)? {
                 if let Some((ge_key, _value)) = &path.curr {
@@ -1433,7 +1480,6 @@ impl Storage {
             }
         }
         path.curr = None;
-        path.stack.clear();
         Ok(false)
     }
 
@@ -1448,20 +1494,22 @@ impl Storage {
             let pin = self.get_page(top.pid, AccessMode::ReadOnly)?;
             let page = self.pool[pin.buf as usize].read().unwrap();
             let n_items = page.get_n_items();
-            if top.pos + inc < n_items {
+            let pos = top.pos + inc;
+            if pos < n_items {
                 path.stack.push(PagePos {
                     pid: top.pid,
-                    pos: top.pos + inc,
+                    pos: pos,
                 });
                 if path.stack.len() == height as usize {
-                    let item = page.get_item(top.pos + inc);
+                    let item = page.get_item(pos);
                     path.curr = Some(item);
                     break;
                 }
                 // We have to use this trick with `inc` variable on the way down because
                 // Rust will detect overflow if we path -1 as pos
+                debug_assert!(page.get_child(pos) != 0);
                 path.stack.push(PagePos {
-                    pid: page.get_child(0),
+                    pid: page.get_child(pos),
                     pos: 0,
                 });
                 inc = 0;
@@ -1482,20 +1530,24 @@ impl Storage {
             let top = path.stack.pop().unwrap();
             let pin = self.get_page(top.pid, AccessMode::ReadOnly)?;
             let page = self.pool[pin.buf as usize].read().unwrap();
-            if top.pos != 0 {
+            let pos = if top.pos == usize::MAX {
+                page.get_n_items()
+            } else {
+                top.pos
+            };
+            if pos != 0 {
                 path.stack.push(PagePos {
                     pid: top.pid,
-                    pos: top.pos - 1,
+                    pos: pos - 1,
                 });
                 if path.stack.len() == height as usize {
-                    let item = page.get_item(top.pos - 1);
+                    let item = page.get_item(pos - 1);
                     path.curr = Some(item);
                     break;
                 }
-                let n_items = page.get_n_items();
                 path.stack.push(PagePos {
-                    pid: page.get_child(n_items - 1),
-                    pos: n_items,
+                    pid: page.get_child(pos - 1),
+                    pos: usize::MAX, // start from last element of the page
                 });
             }
         }
@@ -1521,7 +1573,7 @@ impl Storage {
         to_remove: &mut dyn Iterator<Item = Result<Key>>,
     ) -> Result<()> {
         let mut db = self.db.write().unwrap(); // prevent concurrent access to the database during update operations (MURSIW)
-        assert!(db.state == DatabaseState::Opened);
+        ensure!(db.state == DatabaseState::Opened);
         let mut result = self.do_updates(&mut db, to_upsert, to_remove);
         if result.is_ok() {
             result = self.commit(&mut db);
@@ -1537,12 +1589,26 @@ impl Storage {
     }
 
     ///
-    /// Put (key,value) pair in the storage, it such key already exist, associated value is undated
+    /// Put (key,value) pair in the storage, if such key already exist, associated value is updated
     ///
     pub fn put(&self, key: Key, value: Value) -> Result<()> {
         self.update(&mut iter::once(Ok((key, value))), &mut iter::empty())
     }
 
+    ///
+    /// Store value for u32 key
+    ///
+    pub fn put_u32(&self, key: u32, value: Value) -> Result<()> {
+        self.put(key.to_be_bytes().to_vec(), value)
+    }
+
+    ///
+    /// Store value for u64 key
+    ///
+    pub fn put_u64(&self, key: u64, value: Value) -> Result<()> {
+        self.put(key.to_be_bytes().to_vec(), value)
+    }
+
     ///
     /// Put (key,value) pairs in the storage, exited keys are updated
     ///
@@ -1557,6 +1623,20 @@ impl Storage {
         self.update(&mut iter::empty(), &mut iter::once(Ok(key)))
     }
 
+    ///
+    /// Remove u32 key from the storage, do nothing if not found
+    ///
+    pub fn remove_u32(&self, key: u32) -> Result<()> {
+        self.remove(key.to_be_bytes().to_vec())
+    }
+
+    ///
+    /// Remove u64 key from the storage, do nothing if not found
+    ///
+    pub fn remove_u64(&self, key: u64) -> Result<()> {
+        self.remove(key.to_be_bytes().to_vec())
+    }
+
     ///
     /// Remove keys from the storage, do nothing if not found
     ///
@@ -1574,13 +1654,27 @@ impl Storage {
     }
 
     ///
-    /// Lookup particular key in the storage.
+    /// Lookup key in the storage.
     ///
     pub fn get(&self, key: &Key) -> Result<Option<Value>> {
         let mut iter = self.range((Included(key), Included(key)));
         Ok(iter.next().transpose()?.map(|kv| kv.1))
     }
 
+    ///
+    /// Lookup u32 key in the storage.
+    ///
+    pub fn get_u32(&self, key: u32) -> Result<Option<Value>> {
+        self.get(&key.to_be_bytes().to_vec())
+    }
+
+    ///
+    /// Lookup u64 key in the storage.
+    ///
+    pub fn get_u64(&self, key: u64) -> Result<Option<Value>> {
+        self.get(&key.to_be_bytes().to_vec())
+    }
+
     ///
     /// Returns bidirectional iterator
     ///
@@ -1619,6 +1713,30 @@ impl Storage {
         let db = self.db.read().unwrap();
         db.recovery
     }
+
+    //
+    // Get database info
+    pub fn get_database_info(&self) -> DatabaseInfo {
+        let mut db_info = DatabaseInfo {
+            ..Default::default()
+        };
+        {
+            let db = self.db.read().unwrap();
+            db_info.db_size = db.meta.size as u64 * PAGE_SIZE as u64;
+            db_info.db_used = db.meta.used as u64 * PAGE_SIZE as u64;
+            db_info.tree_height = db.meta.height as usize;
+            db_info.log_size = db.wal_pos;
+            db_info.state = db.state;
+            db_info.n_committed_transactions = db.lsn;
+            db_info.n_aborted_transactions = db.n_aborted_txns;
+        }
+        {
+            let bm = self.buf_mgr.lock().unwrap();
+            db_info.n_cached_pages = bm.cached as usize;
+            db_info.n_pinned_pages = bm.pinned as usize;
+        }
+        db_info
+    }
 }
 
 impl Drop for Storage {
@@ -1632,7 +1750,7 @@ impl<'a> Transaction<'_> {
     /// Commit transaction
     ///
     pub fn commit(&mut self) -> Result<()> {
-        assert!(self.status == TransactionStatus::InProgress);
+        ensure!(self.status == TransactionStatus::InProgress);
         self.storage.commit(&mut self.db)?;
         self.status = TransactionStatus::Committed;
         Ok(())
@@ -1642,31 +1760,59 @@ impl<'a> Transaction<'_> {
     /// Rollback transaction undoing all changes
     ///
     pub fn rollback(&mut self) -> Result<()> {
-        assert!(self.status == TransactionStatus::InProgress);
+        ensure!(self.status == TransactionStatus::InProgress);
         self.storage.rollback(&mut self.db)?;
         self.status = TransactionStatus::Aborted;
         Ok(())
     }
 
     ///
-    /// Insert new key int the storage or update existed key as part of this transaction.
+    /// Insert new key in the storage or update existed key as part of this transaction.
     ///
     pub fn put(&mut self, key: &Key, value: &Value) -> Result<()> {
-        assert!(self.status == TransactionStatus::InProgress);
+        ensure!(self.status == TransactionStatus::InProgress);
         self.storage.do_upsert(&mut self.db, key, value)?;
         Ok(())
     }
 
+    ///
+    /// Store value for u32 key
+    ///
+    pub fn put_u32(&mut self, key: u32, value: &Value) -> Result<()> {
+        self.put(&key.to_be_bytes().to_vec(), value)
+    }
+
+    ///
+    /// Store value for u64 key
+    ///
+    pub fn put_u64(&mut self, key: u64, value: &Value) -> Result<()> {
+        self.put(&key.to_be_bytes().to_vec(), value)
+    }
+
     ///
     /// Remove key from storage as part of this transaction.
     /// Does nothing if key not exist.
     ///
     pub fn remove(&mut self, key: &Key) -> Result<()> {
-        assert!(self.status == TransactionStatus::InProgress);
+        ensure!(self.status == TransactionStatus::InProgress);
         self.storage.do_remove(&mut self.db, key)?;
         Ok(())
     }
 
+    ///
+    /// Remove u32 key from the storage, do nothing if not found
+    ///
+    pub fn remove_u32(&mut self, key: u32) -> Result<()> {
+        self.remove(&key.to_be_bytes().to_vec())
+    }
+
+    ///
+    /// Remove u64 key from the storage, do nothing if not found
+    ///
+    pub fn remove_u64(&mut self, key: u64) -> Result<()> {
+        self.remove(&key.to_be_bytes().to_vec())
+    }
+
     ///
     /// Iterator through pairs in key ascending order.
     /// Byte-wise comparison is used, to it is up to serializer to enforce proper ordering,
@@ -1677,13 +1823,27 @@ impl<'a> Transaction<'_> {
     }
 
     ///
-    /// Lookup particular key in the storage.
+    /// Lookup key in the storage.
     ///
     pub fn get(&self, key: &Key) -> Result<Option<Value>> {
         let mut iter = self.range((Included(key), Included(key)));
         Ok(iter.next().transpose()?.map(|kv| kv.1))
     }
 
+    ///
+    /// Lookup u32 key in the storage.
+    ///
+    pub fn get_u32(&self, key: u32) -> Result<Option<Value>> {
+        self.get(&key.to_be_bytes().to_vec())
+    }
+
+    ///
+    /// Lookup u64 key in the storage.
+    ///
+    pub fn get_u64(&self, key: u64) -> Result<Option<Value>> {
+        self.get(&key.to_be_bytes().to_vec())
+    }
+
     ///
     /// Returns bidirectional iterator
     ///
